/*
 * Parameter tree
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <sys/utsname.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "macros.h"
#include "log.h"
#include "xmalloc.h"
#include "tree.h"
#include "list.h"
#include "buffer.h"
#include "str.h"
#include "vec.h"

#include "par.h"

#define PAR_VERSION_MAJOR	0
#define PAR_VERSION_MINOR	1

/* Each tree node has it's type */
enum {
	NODE_ROOT,	/* */
	NODE_CONTAINER,
	NODE_PARAMETER,
};

static tree_node_t par_tree; /* parameter tree root */

/* tree root */
struct root_info {
	int		node_type;
	char		*name;
	struct utsname	uname;
	int		version;
};

/* simple container */
struct par_container {
	int		node_type;
	char		*name;
};

static int compare_node_names(const void *a, const void *b)
{
	struct par_container *ap, *bp;

	ap = (struct par_container *) a;
	bp = (struct par_container *) b;

	return strcmp(ap->name, bp->name);
}

/* this helper searches the tree in path and returns node or NULL */
struct tree_node *search_for_name(const char *path)
{
	size_t i;
	char *p, *q, *str;
	vector_t vectorized_path;
	struct tree_node *parent, *iter;

	if (*path++ != '/')
		return NULL;

	vectorized_path = v_new();
	str = xstrdup(path);
	p = q = str;

	do {
		if (*p == '/') {
			p++;
			q++;
		} else if (*p) {
			p = strchr(q, '/');
			if (p)
				*p++ = 0;
			if (*q)
				v_add(vectorized_path, q);
			q = p;
		}
	} while (p && *p);

	parent = par_tree;

	for (i = 0; i < v_used(vectorized_path) && parent != NULL; i++) {
		struct par_container tmp;
		tmp.node_type = 0;
		tmp.name = v_ptr(vectorized_path, i);
		iter = tree_node_find_child(parent, &tmp, compare_node_names);
		if (iter)
			parent = iter;
		else
			parent = NULL;
	}

	xfree(str);
	v_free(&vectorized_path);

	return parent;
}

const char *par_type2str(par_type_t type)
{
	switch (type) {
	case PAR_BOOL: return "BOOL";
	case PAR_INT: return "INT";
	case PAR_FLOAT: return "FLOAT";
	case PAR_STRING: return "STRING";
	default: return "ERR";
	}
	return NULL;
}

int par_tree_create()
{
	int err;
	struct root_info *inf;

	if (par_tree)
		error("par_tree_create: tree already exists");

	inf = xmalloc0(sizeof(*inf));
	err = uname(&inf->uname);
	if (err)
		warn("par_tree_create: uname() %d: %s", err, strerror(errno));
	inf->node_type = NODE_ROOT;
	inf->name = "";
	inf->version = PAR_VERSION_MAJOR << 8 | PAR_VERSION_MINOR;
	par_tree = tree_node_new();
	par_tree->data = inf;
	return 0;
}

static int par_tree_node_free(tree_node_t node, void *opaque)
{
	struct par_container *container;
	struct par_struct *par;

	container = (struct par_container *) node->data;

	if (container) {
		if (container->node_type == NODE_CONTAINER) {
			if (container->name)
				xfree(container->name);
			xfree(container);
		} else if (container->node_type == NODE_PARAMETER) {
			par = (struct par_struct *) container;
			if (par->type == PAR_STRING) {
				if (par->value_str != NULL)
					xfree(par->value_str);
				else
					warn("%p PAR_STRING NULL??", par);
			}
			if (par->name)
				xfree(par->name);
			else
				warn("%p PAR name NULL??", par);
			xfree(par);
		} else if (container->node_type == NODE_ROOT) {
			xfree(node->data);
		}
		node->data = NULL;
	} else {
		warn("par_tree_destroy: %p NULL data?", node);
	}
	return 0;
}

void par_tree_destroy()
{
	tree_node_traverse(par_tree, par_tree_node_free, 1, NULL);
	tree_node_free_path(&par_tree);
	par_tree = NULL;
}

int par_parameter_add(struct par_struct **retpar, const char *path,
		const char *name, par_type_t type)
{
	int res = -1;
	int node_type;
	struct par_struct *par;
	tree_node_t node, parent;

	if (!path || !name)
		error("par_parameter_add: path or name NULL");

	if (!par_tree)
		error("par_parameter_add: parent NULL");

	*retpar = NULL;
	parent = search_for_name(path);

	/* Parameter can't be placed in the root.
	   At least one container required. */

	if (parent) {
		node_type = ((struct par_container *)parent->data)->node_type;

		if (node_type != NODE_CONTAINER)
			return PAR_ECONTAINER;

		par = xmalloc(sizeof(*par));
		par->type = type;
		par->quality = PAR_QUAL_INVALID;
		switch (par->type) {
		case PAR_BOOL:
			par->value_bool = 0;
			break;
		case PAR_INT:
			par->value_int = 0;
			break;
		case PAR_FLOAT:
			par->value_float = 0.0;
			break;
		case PAR_STRING:
			par->value_str = xmalloc0(PAR_STR_MAX+1);
			par->value_str[0] = '0';
			break;
		}
		par->node_type = NODE_PARAMETER;
		par->name = xstrdup(name);
		node = tree_node_new();
		node->data = par;
		par->par_node = node;
		tree_node_add(parent, node, -1);
		*retpar = par;
		res = 0;
	} else {
		res = PAR_ENOTFOUND;
	}

	return res;
}

int par_container_add(const char *path, const char *name)
{
	int res = PAR_ENOTFOUND;
	tree_node_t node, parent;
	int node_type;
	struct par_container *container;

	if (!path || !name)
		error("par_container_add: path or name NULL");

	if (!par_tree)
		error("par_container_add: no root");

	parent = search_for_name(path);

	if (parent) {
		node_type = ((struct par_container *)(parent->data))->node_type;
		if (node_type != NODE_ROOT && node_type != NODE_CONTAINER)
			return PAR_ENOTFOUND;
		container = xmalloc0(sizeof(*container));
		container->node_type = NODE_CONTAINER;
		container->name = xstrdup(name);
		node = tree_node_new();
		node->data = container;
		tree_node_add(parent, node, 0);
		res = 0;
	}
	return res;
}

struct par_struct *par_find(const char *full_path_to_parameter)
{
	int node_type;
	struct tree_node *node;
	struct par_struct *par;

	if (!full_path_to_parameter)
		error("par_find: path NULL");

	node = search_for_name(full_path_to_parameter);

	if (node) {
		node_type = ((struct par_container *)node->data)->node_type;
		if (node_type != NODE_PARAMETER)
			return NULL;
		par = node->data;
		return par;
	}
	return NULL;
}

void par_write(struct par_struct *par, par_value_t par_value)
{
	if (!par)
		error("par_write: par NULL");

	switch (par->type) {
	case PAR_BOOL:
		if (par_value.value_bool == PAR_BOOL_ON)
			par->value_bool = PAR_BOOL_ON;
		else if (par_value.value_bool == PAR_BOOL_OFF)
			par->value_bool = PAR_BOOL_OFF;
		break;
	case PAR_INT:
		par->value_int = par_value.value_int;
		break;
	case PAR_FLOAT:
		par->value_float = par_value.value_float;
		break;
	case PAR_STRING:
		if (par_value.value_str != NULL)
			xstrncpy(par->value_str, par_value.value_str, PAR_STR_MAX+1);
		else
			par->value_str[0] = '\0';
		break;
	}
}

par_value_t par_read(struct par_struct *par)
{
	par_value_t value;

	if (!par)
		error("par_read: par NULL");

	switch (par->type) {
	case PAR_BOOL:
		value.value_bool = par->value_bool;
		break;
	case PAR_INT:
		value.value_int = par->value_int;
		break;
	case PAR_FLOAT:
		value.value_float = par->value_float;
		break;
	case PAR_STRING:
		value.value_str = par->value_str;
		break;
	}
	return value;
}

void par_show(struct par_struct *par)
{
	if (par->name)
		printf("'%s' ", par->name);

	switch (par->type) {
	case PAR_BOOL:
		printf("PAR_BOOL state: %s\n", par->value_bool ? "ON" : "OFF");
		break;
	case PAR_INT:
		printf("PAR_INT value: %d\n", par->value_int);
		break;
	case PAR_FLOAT:
		printf("PAR_FLOAT %5.4f\n", par->value_float);
		break;
	case PAR_STRING:
		if (par->value_str != NULL)
			printf("PAR_STRING '%.80s'", par->value_str);
		else
			printf("PAR_STRING NULL");
		break;
	default:
		error("par_show: unknown parameter type");
		break;
	}
}

str_t par_get_full_path(struct par_struct *par)
{
	size_t i;
	char *dir;
	vector_t v;
	tree_node_t iter;
	struct par_container *container;
	str_t str = NULL;

	assert(par->par_node);
	assert(par->par_node->parent);

	v = v_new();
	iter = par->par_node->parent;
	while (iter) {
		container = (struct par_container *) iter->data;
		v_add(v, container->name);
		iter = iter->parent;
	}
	for (i = v_used(v); i > 0; i--) {
		if (!str)
			str = str_new();
		dir = (char *) v_ptr(v, i-1);
		if (*dir) {
			str_append(str, dir);
			str_putc(str, '/');
		} else
			str_append(str, "/");
	}
	if (par->name)
		str_append(str, par->name);
	else
		str_append(str, "/");
	v_free(&v);
	return str;
}

static int par_tree_node_print(tree_node_t node, void *opaque)
{
	size_t i;
	char *dir;
	vector_t v_path;
	tree_node_t iter;
	struct par_container *container;

	v_path = v_new();
	iter = node->parent;
	while (iter) {
		container = (struct par_container *) iter->data;
		v_add(v_path, container->name);
		iter = iter->parent;
	}
	for (i = v_used(v_path); i > 0; i--) {
		dir = (char *) v_ptr(v_path, i-1);
		if (*dir)
			printf("%s/", dir);
		else
			printf("/");
	}
	container = node->data;
	if (*container->name)
		printf("%s\n", container->name);
	else
		printf("/\n");
	v_free(&v_path);
	return 0;
}

void par_tree_list()
{
	tree_node_traverse(par_tree, par_tree_node_print, 1, NULL);
}
